/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file error.h list of error messages for all subsystems */

/** \cond list of all subsystems that support the shiny new error facility */
enum para_subsystem {
	SS_RECV,
	SS_NET,
	SS_ORTP_RECV,
	SS_AUDIOD,
	SS_EXEC,
	SS_CLOSE_ON_FORK,
	SS_SIGNAL,
	SS_STRING,
	SS_DAEMON,
	SS_STAT,
	SS_TIME,
	SS_GRAB_CLIENT,
	SS_HTTP_RECV,
	SS_RECV_COMMON,
	SS_FILTER_CHAIN,
	SS_WAV,
	SS_COMPRESS,
	SS_OGGDEC,
	SS_FILTER,
	SS_COMMAND,
	SS_RANDOM_SELECTOR,
	SS_PLAYLIST_SELECTOR,
	SS_CRYPT,
	SS_HTTP_SEND,
	SS_ORTP_SEND,
	SS_DB,
	SS_OGG,
	SS_MP3,
	SS_MP3DEC,
	SS_SERVER,
	SS_AFS,
	SS_MYSQL_SELECTOR,
	SS_IPC,
	SS_DCCP,
	SS_DCCP_RECV,
	SS_DCCP_SEND,
	SS_FD,
	SS_GUI,
	SS_WRITE,
	SS_WRITE_COMMON,
	SS_ALSA_WRITER,
	SS_FILE_WRITER,
	SS_RINGBUFFER};

#define NUM_SS (SS_RINGBUFFER + 1)
extern const char **para_errlist[];
/** \endcond */

#define NET_ERRORS \
	PARA_ERROR(SEND, "send error"), \
	PARA_ERROR(RECV, "receive error"), \
	PARA_ERROR(SOCKET, "socket error"), \
	PARA_ERROR(CONNECT, "connect error"), \
	PARA_ERROR(ACCEPT, "accept error"), \
	PARA_ERROR(SETSOCKOPT, "failed to set socket options"), \
	PARA_ERROR(BIND, "bind error"), \
	PARA_ERROR(NAME_TOO_LONG, "name too long for struct sockaddr_un"), \
	PARA_ERROR(CHMOD, "failed to set socket mode"), \
	PARA_ERROR(SENDMSG, "sendmsg() failed"), \
	PARA_ERROR(RECVMSG, "recvmsg() failed"), \
	PARA_ERROR(SCM_CREDENTIALS, "did not receive SCM credentials"), \
	PARA_ERROR(LISTEN, "listen error"), \
	PARA_ERROR(RECV_PATTERN, "did not receive expected pattern"), \
	PARA_ERROR(HOST_INFO, "gethostbyname() failed"), \


#define ORTP_RECV_ERRORS \
	PARA_ERROR(MSG_TO_BUF, "failed to extract rtp packet"), \
	PARA_ERROR(ORTP_SYNTAX, "ottp syntax error"), \
	PARA_ERROR(TOO_MANY_BAD_CHUNKS, "too many consecutive bad chunks"), \
	PARA_ERROR(INVALID_HEADER, "invalid header packet"), \
	PARA_ERROR(OVERRUN, "outout buffer overrun"), \


#define HTTP_RECV_ERRORS \
	PARA_ERROR(SEND_HTTP_REQUEST, "failed to send http request"), \
	PARA_ERROR(MISSING_OK, "did not receive OK message from peer"), \
	PARA_ERROR(HTTP_RECV_BUF, "did not receive buffer")


#define RECV_ERRORS \
	PARA_ERROR(RECV_SELECT, "recv select error"), \
	PARA_ERROR(WRITE_STDOUT, "stdout write error"), \


#define RECV_COMMON_ERRORS \
	PARA_ERROR(RECV_SYNTAX, "recv syntax error"), \


#define AUDIOD_ERRORS \
	PARA_ERROR(WRITE_AUDIO_DATA, "failed to write audio data"), \
	PARA_ERROR(NO_MORE_SLOTS, "no more empty slots"), \
	PARA_ERROR(MISSING_COLON, "syntax error: missing colon"), \
	PARA_ERROR(UNSUPPORTED_AUDIO_FORMAT, "given audio format not supported"), \
	PARA_ERROR(CLIENT_WRITE, "client write error"), \
	PARA_ERROR(UCRED_PERM, "permission denied"), \
	PARA_ERROR(INVALID_AUDIOD_CMD, "invalid command"), \


#define FILTER_CHAIN_ERRORS \
	PARA_ERROR(UNSUPPORTED_FILTER, "given filter not supported"), \
	PARA_ERROR(BAD_FILTER_OPTIONS, "invalid filter option given"), \


#define STAT_ERRORS \
	PARA_ERROR(TOO_MANY_CLIENTS, "maximal number of stat clients exceeded"), \
	PARA_ERROR(UNKNOWN_STAT_ITEM, "status item not recognized"), \


#define OGGDEC_ERRORS \
	PARA_ERROR(OGGDEC_READ, "read from media returned an error"), \
	PARA_ERROR(OGGDEC_NOTVORBIS, "bitstream is not vorbis data"), \
	PARA_ERROR(OGGDEC_VERSION, "vorbis version mismatch"), \
	PARA_ERROR(OGGDEC_BADHEADER, "invalid vorbis bitstream header"), \
	PARA_ERROR(OGGDEC_FAULT, "bug or heap/stack corruption"), \
	PARA_ERROR(OGGDEC_BADLINK, "invalid stream section or requested link corrupt"), \


#define GRAB_CLIENT_ERRORS \
	PARA_ERROR(PEDANTIC_GRAB, "fd not ready and pedantic grab requested"), \
	PARA_ERROR(GC_WRITE, "grab client write error"), \
	PARA_ERROR(BAD_GC_SLOT, "invalid slot requested by grab client"), \
	PARA_ERROR(BAD_GC_FILTER_NUM, "invalid filter number given"), \
	PARA_ERROR(GC_SYNTAX, "grab client syntax error"), \
	PARA_ERROR(GC_HELP_GIVEN, ""), /* not really an error */ \
	PARA_ERROR(GC_VERSION_GIVEN, ""), /* not really an error */ \


#define MP3DEC_ERRORS \
	PARA_ERROR(MAD_FRAME_DECODE, "mad frame decode error"), \
	PARA_ERROR(MP3DEC_OVERRUN, "mp3 output buffer overrun"), \


#define FILTER_ERRORS \
	PARA_ERROR(NO_FILTERS, "at least one filter must be given"), \
	PARA_ERROR(FILTER_SYNTAX, "syntax error"), \


#define SIGNAL_ERRORS \
	PARA_ERROR(SIGNAL_SIG_ERR, "signal() retured SIG_ERR"), \
	PARA_ERROR(SIGNAL_READ, "read error from signal pipe"), \
	PARA_ERROR(WAITPID, "waitpid error"), \
	PARA_ERROR(SIGNAL_PIPE, "failed to setup signal pipe"), \


#define STRING_ERRORS \
	PARA_ERROR(MKSTEMP, "mkstemp error: unable to create tmp file"), \
	PARA_ERROR(FCHMOD, "fchmod error: can not set mode"), \


#define EXEC_ERRORS \
	PARA_ERROR(DUP_PIPE, "exec error: can not create pipe"), \
	PARA_ERROR(NULL_OPEN, "can not open /dev/null"), \


#define MP3_ERRORS \
	PARA_ERROR(FREAD, "fread error"), \
	PARA_ERROR(FSEEK, "fseek error"), \
	PARA_ERROR(FRAME, "invalid mp3 frame"), \
	PARA_ERROR(FRAME_LENGTH, "invalid frame length"), \
	PARA_ERROR(MP3_NO_FILE, "invalid mp3 file pointer"), \
	PARA_ERROR(MP3_INFO, "could not read mp3 info"), \
	PARA_ERROR(MP3_REPOS, "mp3 repositioning error"), \
	PARA_ERROR(HEADER_FREQ, "invalid header frequency"), \
	PARA_ERROR(HEADER_BITRATE, "invalid header bitrate"), \


#define OGG_ERRORS \
	PARA_ERROR(OGG_READ, "ogg read error"), \
	PARA_ERROR(SYNC_PAGEOUT, "ogg sync page-out error (no ogg file?)"), \
	PARA_ERROR(STREAM_PAGEIN, "ogg stream page-in error (first page)"), \
	PARA_ERROR(STREAM_PACKETOUT, "ogg stream packet-out error (first packet)"), \
	PARA_ERROR(VORBIS, "vorbis synthesis header-in error (not vorbis?)"), \
	PARA_ERROR(OGG_NO_FILE, "invalid ogg file pointer"), \
	PARA_ERROR(OGG_OPEN, "ov_open error"), \
	PARA_ERROR(OGG_INFO, "ov_info error"), \
	PARA_ERROR(OGG_REPOS, "ogg repositioning error"), \


#define AFS_ERRORS \
	PARA_ERROR(AUDIO_FORMAT, "audio format not recognized"), \
	PARA_ERROR(FSTAT, "failed to fstat() audio file"), \


#define DB_ERRORS \
	PARA_ERROR(GETCWD, "can not get current working directory"), \
	PARA_ERROR(CHDIR, "can not change directory"), \
	PARA_ERROR(OPENDIR, "can not open directory"), \
	PARA_ERROR(LSTAT, "lstat error"), \


#define CRYPT_ERRORS \
	PARA_ERROR(PRIVATE_KEY, "can not read private key"), \
	PARA_ERROR(PUBLIC_KEY, "can not read public key"), \
	PARA_ERROR(RSA, "RSA error"), \
	PARA_ERROR(ENCRYPT, "encrypt error"), \
	PARA_ERROR(DECRYPT, "decrypt error"), \
	PARA_ERROR(CHALLENGE, "failed to read challenge"), \


#define HTTP_SEND_ERRORS \
	PARA_ERROR(QUEUE, "packet queue overrun"), \
	PARA_ERROR(WRITE_OK, "can not check whether fd is writable"), \


#define RANDOM_SELECTOR_ERRORS \
	PARA_ERROR(FILE_COUNT, "audio file count exceeded"), \
	PARA_ERROR(NOTHING_FOUND, "no audio files found"), \


#define MYSQL_SELECTOR_ERRORS \
	PARA_ERROR(MYSQL_SYNTAX, "mysql syntax error"), \
	PARA_ERROR(NOTCONN, "not connected to mysql server"), \
	PARA_ERROR(TOOBIG, "mysql: file too large"), \
	PARA_ERROR(NAMETOOLONG, "mysql: name too long"), \
	PARA_ERROR(QFAILED, "mysql query failed"), \
	PARA_ERROR(NOROW, "row is NULL"), \
	PARA_ERROR(NOATTS, "can not get attributes from mysql table"), \
	PARA_ERROR(NORESULT, "error while fetching mysql result"), \
	PARA_ERROR(EMPTY_RESULT, "result is empty"), \
	PARA_ERROR(ESCAPE, "can not escape string"), \
	PARA_ERROR(GET_AUDIO_FILE, "can not get current audio file"), \
	PARA_ERROR(GET_STREAM, "can not get current stream"), \
	PARA_ERROR(NO_STREAM, "no such stream"), \
	PARA_ERROR(AUDIO_FILE, "no such audio file"), \
	PARA_ERROR(GET_QUERY, "can not get query for specified stream"), \
	PARA_ERROR(TMPFILE, "error while writing temporary file"), \
	PARA_ERROR(META, "can not get meta data"), \
	PARA_ERROR(MYSQL_INIT, "can not initialize mysql connection"), \
	PARA_ERROR(NO_MYSQL_PASSWD, "fatal: no mysql passord given"), \
	PARA_ERROR(NO_AF_DIR, "fatal: audio file directory not given"), \


#define COMMAND_ERRORS \
	PARA_ERROR(COMMAND_SYNTAX, "syntax error in command"), \
	PARA_ERROR(AUTH, "did not receive auth request"), \
	PARA_ERROR(BAD_SELECTOR, "no such audio file selector"), \
	PARA_ERROR(NO_AUDIO_FILE, "no audio file"), \
	PARA_ERROR(BAD_CMD, "invalid command"), \
	PARA_ERROR(PERM, "permission denied"), \
	PARA_ERROR(USERLIST, "failed to open user list file"), \
	PARA_ERROR(BAD_USER, "you don't exist. Go away."), \
	PARA_ERROR(LOCK, "lock error"), \
	PARA_ERROR(SENDER_CMD, "command not supported by this sender"), \

#define PLAYLIST_SELECTOR_ERRORS \
	PARA_ERROR(LOAD_PLAYLIST, "failed to load playlist"), \


#define IPC_ERRORS \
	PARA_ERROR(SEM_GET, "failed to create semaphore"), \
	PARA_ERROR(SEM_REMOVE, "can not remove semaphore"), \
	PARA_ERROR(SHM_GET, "failed to allocate shared memory area"), \
	PARA_ERROR(SHM_DESTROY, "failed to destroy shared memory area"), \
	PARA_ERROR(SHM_ATTACH, "can not attach shared memory area"), \
	PARA_ERROR(SHM_DETACH, "can not detach shared memory area"), \


#define DCCP_ERRORS \
	PARA_ERROR(DCCP_SOCKET, "can not create dccp socket"), \
	PARA_ERROR(DCCP_PACKET_SIZE, "failed to set dccp packet size"), \
	PARA_ERROR(DCCP_SERVICE, "could not get service code"), \


#define DCCP_RECV_ERRORS \
	PARA_ERROR(ADDR_INFO, "getaddrinfo error"), \
	PARA_ERROR(DCCP_OVERRUN, "dccp output buffer buffer overrun"), \
	PARA_ERROR(DCCP_CONNECT, "dccp connect error"), \

#define DCCP_SEND_ERRORS \
	PARA_ERROR(DCCP_BIND, "dccp bind error"), \
	PARA_ERROR(DCCP_LISTEN, "dccp listen error"), \
	PARA_ERROR(DCCP_WRITE, "dccp write error"), \

#define FD_ERRORS \
	PARA_ERROR(F_GETFL, "failed to get fd flags"), \
	PARA_ERROR(F_SETFL, "failed to set fd flags"), \


#define WRITE_ERRORS \
	PARA_ERROR(READ_HDR, "failed to read audio file header"), \
	PARA_ERROR(READ_STDIN, "failed to read from stdin"), \
	PARA_ERROR(WRITE_SYNTAX, "para_write syntax error"), \
	PARA_ERROR(WRITE_OVERRUN, "buffer overrun"), \
	PARA_ERROR(PREMATURE_END, "premature end of audio file"), \


#define ALSA_WRITER_ERRORS \
	PARA_ERROR(BROKEN_CONF, "Broken alsa configuration"), \
	PARA_ERROR(ACCESS_TYPE, "alsa access type not available"), \
	PARA_ERROR(SAMPLE_FORMAT, "sample format not available"), \
	PARA_ERROR(CHANNEL_COUNT, "channels count not available"), \
	PARA_ERROR(HW_PARAMS, "unable to install hw params"), \
	PARA_ERROR(SW_PARAMS, "unable to install sw params"), \
	PARA_ERROR(BAD_PERIOD, "can not use period equal to buffer size"), \
	PARA_ERROR(GET_XFER, "unable to obtain xfer align"), \
	PARA_ERROR(SET_XFER, "snd_pcm_sw_params_set_xfer_align() failed"), \
	PARA_ERROR(ALSA_WRITE, "alsa write error"), \
	PARA_ERROR(PCM_OPEN, "unable to open pcm"), \
	PARA_ERROR(SND_PCM_INFO, "pcm info error"), \
	PARA_ERROR(GET_BUFFER_TIME, "snd_pcm_hw_params_get_buffer_time_max() failed"), \
	PARA_ERROR(SET_BUFFER_TIME, "snd_pcm_hw_params_set_buffer_time_near() failed"), \
	PARA_ERROR(SET_RATE, "snd_pcm_hw_params_set_rate_near failed"), \
	PARA_ERROR(START_THRESHOLD, "snd_pcm_sw_params_set_start_threshold() failed"), \
	PARA_ERROR(STOP_THRESHOLD, "snd_pcm_sw_params_set_stop_threshold() failed"), \
	PARA_ERROR(ALSA_LOG, "snd_output_stdio_attach() failed"), \


#define FILE_WRITER_ERRORS \
	PARA_ERROR(FW_WRITE, "file writer write error"), \
	PARA_ERROR(FW_OPEN, "file writer: can not open output file"), \

#define WRITE_COMMON_ERRORS \
	PARA_ERROR(WRITE_COMMON_SYNTAX, "syntax error in write option"), \


/* these do not need error handling (yet) */
#define SERVER_ERRORS
#define WAV_ERRORS
#define COMPRESS_ERRORS
#define TIME_ERRORS
#define CLOSE_ON_FORK_ERRORS
#define DAEMON_ERRORS
#define ORTP_SEND_ERRORS
#define GUI_ERRORS
#define RINGBUFFER_ERRORS


/**
 * the subsystem shift
 *
 * 255 error codes ought to be enough for every subsystem. Use the higher bits
 * of the return value to encode the subsystem number
 */
#define SS_SHIFT 8

/**
 * compute the subsystem offset
 *
 * It is given by x * 2**8 where \a x is the subsystem number
 */
#define SS_OFFSET(ss) (SS_ ## ss << SS_SHIFT)

/**
 * make the enum of all errors of one subsystem
 *
 * As zero should not be an error, we define a dummy enum entry with value
 * 2**ss. That lets the real errors start at 2**ss + 1.
 */
#define SS_ENUM(ss) enum {\
	E_ ## ss ## _DUMMY = SS_OFFSET(ss), \
	ss ## _ERRORS}

/**
 * determine the subsystem number from the error number
 *
 * Easy, it's just \a num / 2**8.
 */
#define ERRNUM_TO_SS(num) ((num) >> SS_SHIFT)

/**
 * determine the index of an error number
 *
 * Also easy: It's the lower 8 bits of num - 1.
 */
#define ERRNUM_TO_INDEX(num) (((1 << SS_SHIFT) - 1) & ((num) - 1))

/**
 * paraslash's version of strerror(3)
 *
 * expands to the error text of \a num (a string constant).
 */
#define PARA_STRERROR(num) para_errlist[ERRNUM_TO_SS(num)] [ERRNUM_TO_INDEX(num)]

/**
 * define the error list for one subsystem
 *
 * Used by macros in config.h (generated by configure)
 */
#define DEFINE_ERRLIST(ss) const char * ss ## _ERRLIST[] = {ss ## _ERRORS}

/**
 * activate errors for one subsystem.
 *
 * Each executable needs only the error lists of those subssystems it is actually
 * linked with. We always reserve zeroed-out space for NUM_SS char ** pointers, but
 * only init those of the needed subsystems. This macro is used by macros in config.h
 * (generated by configure).
 */
#define PARA_ERRLIST(ss) [SS_ ## ss] = ss ## _ERRLIST

/**
 * This is temporarily defined to expand to its second argument (prefixed by
 * 'E_') and gets later redefined to expand to the error text only
 */
#define PARA_ERROR(err, msg) E_ ## err

/** \cond popcorn time */
SS_ENUM(ORTP_RECV);
SS_ENUM(NET);
SS_ENUM(RECV);
SS_ENUM(AUDIOD);
SS_ENUM(EXEC);
SS_ENUM(CLOSE_ON_FORK);
SS_ENUM(SIGNAL);
SS_ENUM(STRING);
SS_ENUM(DAEMON);
SS_ENUM(STAT);
SS_ENUM(TIME);
SS_ENUM(GRAB_CLIENT);
SS_ENUM(HTTP_RECV);
SS_ENUM(RECV_COMMON);
SS_ENUM(FILTER_CHAIN);
SS_ENUM(WAV);
SS_ENUM(COMPRESS);
SS_ENUM(OGGDEC);
SS_ENUM(MP3DEC);
SS_ENUM(FILTER);
SS_ENUM(MP3);
SS_ENUM(OGG);
SS_ENUM(SERVER);
SS_ENUM(AFS);
SS_ENUM(COMMAND);
SS_ENUM(RANDOM_SELECTOR);
SS_ENUM(PLAYLIST_SELECTOR);
SS_ENUM(CRYPT);
SS_ENUM(HTTP_SEND);
SS_ENUM(ORTP_SEND);
SS_ENUM(DB);
SS_ENUM(MYSQL_SELECTOR);
SS_ENUM(IPC);
SS_ENUM(DCCP);
SS_ENUM(DCCP_RECV);
SS_ENUM(DCCP_SEND);
SS_ENUM(FD);
SS_ENUM(GUI);
SS_ENUM(WRITE);
SS_ENUM(WRITE_COMMON);
SS_ENUM(ALSA_WRITER);
SS_ENUM(FILE_WRITER);
SS_ENUM(RINGBUFFER);
/** \endcond */
#undef PARA_ERROR
/* rest of the world only sees the error text */
#define PARA_ERROR(err, msg) msg
